import os
import json
import re
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import (
    DomainModel, IntegerType, StringType, Class,
    BooleanType, FloatType, Enumeration, DateType,
    DateTimeType, TimeType,
)
from besser.generators import GeneratorInterface

class JSONSchemaGenerator(GeneratorInterface):
    """
    JSONSchemaGenerator is a class that implements the GeneratorInterface and is responsible for generating
    JSONSchema code based on B-UML models.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
        mode (str, optional): The generation mode, either 'regular' or 'smart_data'. Defaults to 'regular'.
    """

    def __init__(self, model: DomainModel, output_dir: str = None, mode: str = 'regular'):
        super().__init__(model, output_dir)
        self.mode = mode
        # Add enums to TYPES dictionary

    def _get_property_type(self, property_type):
        """
        Maps B-UML types to JSON schema types and returns format when needed.
        
        Args:
            property_type: The B-UML type to map.
            
        Returns:
            tuple: (type, format) where format is None if not applicable
        """
        # Check if it's an enumeration instance
        if hasattr(property_type, '__class__') and property_type.__class__.__name__ == 'Enumeration':
            return "string", None

        # For BUML types, they are usually type objects/classes, not instances
        # Check the actual type or class name
        if property_type == IntegerType or (hasattr(property_type, '__name__') and property_type.__name__ == 'IntegerType'):
            return "integer", None
        elif property_type == StringType or (hasattr(property_type, '__name__') and property_type.__name__ == 'StringType'):
            return "string", None
        elif property_type == BooleanType or (hasattr(property_type, '__name__') and property_type.__name__ == 'BooleanType'):
            return "boolean", None
        elif property_type == FloatType or (hasattr(property_type, '__name__') and property_type.__name__ == 'FloatType'):
            return "number", None
        elif property_type == DateTimeType or (hasattr(property_type, '__name__') and property_type.__name__ == 'DateTimeType'):
            return "string", "date-time"
        elif property_type == DateType or (hasattr(property_type, '__name__') and property_type.__name__ == 'DateType'):
            return "string", "date"
        elif property_type == TimeType or (hasattr(property_type, '__name__') and property_type.__name__ == 'TimeType'):
            return "string", "time"
        else:
            return "string", None

    def _get_property_description(self, attr, prop_type):
        """
        Generate proper description format for Smart Data Models.
        
        Args:
            attr: The attribute object
            prop_type: The JSON schema type
            
        Returns:
            str: Formatted description
        """
        # Map JSON schema types to schema.org URLs
        schema_org_mapping = {
            "integer": "https://schema.org/Number",
            "number": "https://schema.org/Number", 
            "string": "https://schema.org/Text",
            "boolean": "https://schema.org/Boolean"
        }
        
        model_url = schema_org_mapping.get(prop_type, "https://schema.org/Text")
        
        if attr.metadata and attr.metadata.description:
            return f"Property. Model:'{model_url}'. {attr.metadata.description}"
        else:
            return f"Property. Model:'{model_url}'. {attr.name} value"

    def _prepare_smart_data_schema_for_class(self, class_def):
        """
        Prepares schema data for Smart Data format for a specific class.
        
        Args:
            class_def: The class to prepare the schema for.
            
        Returns:
            dict: A dictionary containing the schema data.
        """
        # Build class-specific properties
        class_properties = {}
        
        # Add the mandatory type property
        class_properties["type"] = {
            "type": "string",
            "enum": [class_def.name],
            "description": "Property. NGSI Entity type"
        }
        
        # Process class attributes
        for attr in class_def.attributes:
            prop_type, prop_format = self._get_property_type(attr.type)
            prop_def = {"type": prop_type}

            if prop_format is not None:
                prop_def["format"] = prop_format

            # Use the proper description format
            prop_def["description"] = self._get_property_description(attr, prop_type)

            if hasattr(attr.type, '__class__') and attr.type.__class__.__name__ == 'Enumeration':
                prop_def["enum"] = [lit.name for lit in attr.type.literals]
                prop_def["type"] = "string"

            if hasattr(attr, 'multiplicity') and attr.multiplicity.max > 1:
                prop_def = {
                    "type": "array",
                    "items": prop_def
                }

            class_properties[attr.name] = prop_def

        # Process association ends for this class
        for association in self.model.associations:
            class_ends = [end for end in association.ends if end.type == class_def]

            if not class_ends:
                continue

            for class_end in class_ends:
                other_ends = [end for end in association.ends if end != class_end]
                if not other_ends:
                    continue

                other_end = other_ends[0]
                prop_name = other_end.name

                if not prop_name:
                    continue

                relationship_description = (
                    f"Relationship. Model:'https://schema.org/URL'. Reference to {other_end.type.name}"
                    if association.metadata and association.metadata.description
                    else f"Relationship. Model:'https://schema.org/URL'. Reference to {other_end.type.name}"
                )

                relationship_def = {
                    "description": relationship_description,
                    "type": "string",
                    "format": "uri"
                }

                # Handle multiplicity
                if hasattr(other_end, 'multiplicity') and other_end.multiplicity:
                    if other_end.multiplicity.max > 1:
                        class_properties[prop_name] = {
                            "type": "array",
                            "items": relationship_def
                        }
                    else:
                        class_properties[prop_name] = relationship_def
                else:
                    class_properties[prop_name] = relationship_def

        # Create the complete schema structure
        schema_data = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "$schemaVersion": "0.0.1",
            "$id": f"https://smart-data-models.github.io/dataModel.{self.model.name}/{class_def.name}/schema.json",
            "title": f"Smart Data models - {class_def.name} schema",
            "modelTags": "",
            "description": (
                class_def.metadata.description if class_def.metadata and class_def.metadata.description 
                else f"This class represents {class_def.name} for smart data models implementation"
            ),
            "type": "object",
            "required": ["id", "type"],
            "allOf": [
                {
                    "$ref": "https://smart-data-models.github.io/data-models/common-schema.json#/definitions/GSMA-Commons"
                },
                {
                    "$ref": "https://smart-data-models.github.io/data-models/common-schema.json#/definitions/Location-Commons"
                },
                {
                    "properties": class_properties
                }
            ],
            "derivedFrom": "3GPP TS 28.541",
            "license": ""
        }

        return schema_data

    def generate(self):
        """
        Generates JSONSchema code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file
        """
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))

        env.filters['tojson_ordered'] = self.tojson_ordered

        if self.mode == 'smart_data':
            # Smart Data mode - generate a schema for each class
            template = env.get_template('smart_data_schema.json.j2')

            # Get all classes from the model
            classes = [c for c in self.model.types if isinstance(c, Class)]

            for class_def in classes:
                # Create a directory for each class
                class_dir = os.path.join(self.output_dir, class_def.name)
                os.makedirs(class_dir, exist_ok=True)

                # Create examples directory
                examples_dir = os.path.join(class_dir, "examples")
                os.makedirs(examples_dir, exist_ok=True)

                # Prepare schema data for this class
                schema_data = self._prepare_smart_data_schema_for_class(class_def)
                
                # Generate schema.json file
                file_path = os.path.join(class_dir, "schema.json")
                with open(file_path, mode="w", encoding='utf-8') as f:
                    generated_code = template.render(schema=schema_data)
                    f.write(generated_code)

                # Generate all required example files
                base_example = self.generate_base_example(class_def)
                
                # example.json (key-value format)
                example_path = os.path.join(examples_dir, "example.json")
                with open(example_path, mode="w", encoding='utf-8') as f:
                    f.write(json.dumps(base_example, indent=2))

                # example-normalized.json (NGSI v2 normalized format)
                normalized_example = self.generate_normalized_example(base_example)
                normalized_path = os.path.join(examples_dir, "example-normalized.json")
                with open(normalized_path, mode="w", encoding='utf-8') as f:
                    f.write(json.dumps(normalized_example, indent=2))

                # example.jsonld (JSON-LD format)
                jsonld_example = self.generate_jsonld_example(base_example)
                jsonld_path = os.path.join(examples_dir, "example.jsonld")
                with open(jsonld_path, mode="w", encoding='utf-8') as f:
                    f.write(json.dumps(jsonld_example, indent=2))

                # example-normalized.jsonld (NGSI-LD normalized format)
                normalized_jsonld_example = self.generate_normalized_jsonld_example(base_example)
                normalized_jsonld_path = os.path.join(examples_dir, "example-normalized.jsonld")
                with open(normalized_jsonld_path, mode="w", encoding='utf-8') as f:
                    f.write(json.dumps(normalized_jsonld_example, indent=2))

                # Create ADOPTERS.yaml and notes.yaml in each class directory
                adopters_file_path = os.path.join(class_dir, "ADOPTERS.yaml")
                notes_file_path = os.path.join(class_dir, "notes.yaml")

                with open(adopters_file_path, mode="w", encoding='utf-8') as f:
                    f.write("# List of adopters\n")

                with open(notes_file_path, mode="w", encoding='utf-8') as f:
                    f.write("# Notes about the data model\n")

                print(f"Smart Data schema for {class_def.name} generated in: {file_path}")
        else:
            # Regular JSON Schema mode
            template = env.get_template('json_schema.json.j2')
            file_path = self.build_generation_path(file_name="json_schema.json")

            with open(file_path, mode="w") as f:
                generated_code = template.render(
                    classes=self.model.classes_sorted_by_inheritance(),
                    enumerations = self.model.get_enumerations(),
                )
                f.write(generated_code)
                print("Code generated in the location: " + file_path)

    def generate_example(self, schema_str):
        """Generates an example JSON based on the rendered schema string."""
        schema = json.loads(schema_str)

        title = schema.get('title', 'Entity')
        match = re.search(r'Smart Data models - (\w+)', title)
        model_name = match.group(1) if match else title

        # Extract all properties
        all_props = {}
        if 'properties' in schema:
            all_props = schema['properties']
        else:
            for block in schema.get('allOf', []):
                if 'properties' in block:
                    all_props.update(block['properties'])

        # Build NGSI-LD example
        example_obj = {
            "id": f"urn:ngsi-ld:{model_name}:{model_name}-001",
            "type": model_name
        }

        for prop, details in all_props.items():
            ngsi_type, value = extract_example_and_type(prop, details)
            example_obj[prop] = {
                "type": ngsi_type,
                "value": value
            }

        return json.dumps(example_obj, indent=4)

    def generate_base_example(self, class_def):
        """Generate a base example in key-value format."""
        example = {
            "id": f"urn:ngsi-ld:{class_def.name}:{class_def.name}-001",
            "type": class_def.name
        }

        # Add common Smart Data Model properties
        example.update({
            "name": "example-name",
            "description": "example-description", 
            "dateCreated": "2025-06-25T08:00:00Z",
            "dateModified": "2025-06-30T14:15:00Z",
            "source": "https://example.org/source"
        })

        # Add class attributes
        for attr in class_def.attributes:
            example[attr.name] = self._get_example_value_for_attribute(attr)

        # Add association relationships
        for association in self.model.associations:
            class_ends = [end for end in association.ends if end.type == class_def]
            if not class_ends:
                continue

            for class_end in class_ends:
                other_ends = [end for end in association.ends if end != class_end]
                if not other_ends:
                    continue

                other_end = other_ends[0]
                if other_end.name:
                    example[other_end.name] = f"urn:ngsi-ld:{other_end.name}:{other_end.name}-01"

        return example

    def generate_normalized_example(self, base_example):
        """Generate NGSI v2 normalized example."""
        normalized = {}
        
        for key, value in base_example.items():
            if key in ["id", "type"]:
                normalized[key] = value
            else:
                # Determine the NGSI type based on the value
                if isinstance(value, str) and value.startswith("urn:ngsi-ld:"):
                    normalized[key] = {
                        "type": "Relationship",
                        "value": value
                    }
                elif isinstance(value, str) and ("date" in key.lower() or "time" in key.lower()):
                    normalized[key] = {
                        "type": "DateTime", 
                        "value": value
                    }
                elif isinstance(value, bool):
                    normalized[key] = {
                        "type": "Boolean",
                        "value": value
                    }
                elif isinstance(value, (int, float)):
                    normalized[key] = {
                        "type": "Number",
                        "value": value
                    }
                else:
                    normalized[key] = {
                        "type": "Text",
                        "value": value
                    }

        return normalized

    def generate_jsonld_example(self, base_example):
        """Generate JSON-LD example."""
        jsonld = base_example.copy()
        jsonld["@context"] = [
            "https://smartdatamodels.org/context.jsonld"
        ]
        return jsonld

    def generate_normalized_jsonld_example(self, base_example):
        """Generate NGSI-LD normalized example."""
        normalized_jsonld = {}
        
        for key, value in base_example.items():
            if key in ["id", "type"]:
                normalized_jsonld[key] = value
            else:
                # Create NGSI-LD format
                if isinstance(value, str) and value.startswith("urn:ngsi-ld:"):
                    normalized_jsonld[key] = {
                        "type": "Relationship",
                        "object": value
                    }
                elif isinstance(value, str) and ("date" in key.lower() or "time" in key.lower()):
                    normalized_jsonld[key] = {
                        "type": "Property",
                        "value": {
                            "@type": "DateTime",
                            "@value": value
                        }
                    }
                else:
                    normalized_jsonld[key] = {
                        "type": "Property",
                        "value": value
                    }

        normalized_jsonld["@context"] = [
            "https://smartdatamodels.org/context.jsonld"
        ]
        
        return normalized_jsonld

    def _get_example_value_for_attribute(self, attr):
        """Get example value for a specific attribute."""
        if hasattr(attr.type, '__class__') and attr.type.__class__.__name__ == 'Enumeration':
            return next(iter(attr.type.literals)).name if attr.type.literals else "enum-value"
        elif attr.type == IntegerType or (hasattr(attr.type, '__name__') and attr.type.__name__ == 'IntegerType'):
            return 10
        elif attr.type == FloatType or (hasattr(attr.type, '__name__') and attr.type.__name__ == 'FloatType'):
            return 3.14
        elif attr.type == BooleanType or (hasattr(attr.type, '__name__') and attr.type.__name__ == 'BooleanType'):
            return True
        elif attr.type == DateTimeType or (hasattr(attr.type, '__name__') and attr.type.__name__ == 'DateTimeType'):
            return "2025-06-25T08:00:00Z"
        elif attr.type == DateType or (hasattr(attr.type, '__name__') and attr.type.__name__ == 'DateType'):
            return "2025-06-25"
        elif attr.type == TimeType or (hasattr(attr.type, '__name__') and attr.type.__name__ == 'TimeType'):
            return "08:00:00Z"
        else:  # StringType or default
            return f"example-{attr.name}"

    @staticmethod
    def tojson_ordered(value, indent=4, base_indent=0):
        """
        Serializa JSON ordenado con indentación adicional solo para líneas internas,
        manteniendo la primera línea sin sangría extra.

        Args:
            value (dict or list): La estructura JSON a serializar.
            indent (int): Espacios de indentación por nivel.
            base_indent (int): Espacios extra agregados a todas las líneas excepto la primera.

        Returns:
            str: Cadena JSON con indentación formateada correctamente.
        """
        json_str = json.dumps(value, indent=indent)
        lines = json_str.splitlines()

        if base_indent > 0 and len(lines) > 1:
            pad = ' ' * base_indent
            lines = [lines[0]] + [pad + line for line in lines[1:]]

        return '\n'.join(lines)

def extract_example_and_type(prop, details):
    """Infer NGSI type and value based on JSON schema, handling anyOf, array, and formats."""

    # anyOf: look for 'format' or 'type'
    if 'anyOf' in details:
        option = details['anyOf'][-1]  # Take the last option only
        ngsi_type, value = extract_example_and_type(prop, option)
        return ngsi_type, value

    # array: process items
    if details.get('type') == 'array':
        items = details.get('items', {})
        ngsi_type, item_value = extract_example_and_type(prop, items)
        return ngsi_type, [item_value]

    # Use type and format to determine NGSI type
    ngsi_type = get_ngsi_type(details)
    value = get_example_value(prop, details)

    return ngsi_type, value

def get_example_value(prop, details):
    """Returns a mock example value based on the schema property definition."""

    # Handle anyOf by recursively checking the first valid option
    if 'anyOf' in details:
        for option in details['anyOf']:
            val = get_example_value(prop, option)
            if val is not None:
                return val

    # Handle arrays by generating example for each item
    if details.get('type') == 'array':
        items = details.get('items', {})
        item_val = get_example_value(prop, items)
        return [item_val]

    t = details.get('type')
    fmt = details.get('format')
    enum = details.get('enum')

    if enum:
        return enum[0]

    # Generate values based on type/format
    if t == 'string':
        if fmt == 'uri':
            return f"urn:ngsi-ld:{prop}:{prop}-001"
        elif fmt == 'date':
            return "2023-10-01"
        elif fmt == 'date-time':
            return "2023-10-01T12:00:00Z"
        elif fmt == 'time':
            return "14:30:00Z"
        else:
            return f"example-{prop}"

    if t == 'integer':
        return 30

    if t == 'number':
        return 30 if "money" in prop.lower() else 3.14

    if t == 'boolean':
        return True

    if t == 'object':
        return {"exampleKey": "exampleValue"}

    return f"example-{prop}"

def get_ngsi_type(details):
    """Determines the NGSI type label based on JSON Schema field definition."""

    t = details.get('type')
    fmt = details.get('format')

    # Direct relationship: string + uri format
    if t == 'string' and fmt == 'uri':
        return 'Relationship'

    # Arrays may contain relationships via anyOf in items
    if t == 'array':
        items = details.get('items', {})

        # If items has anyOf, check if any option is uri
        if isinstance(items, dict) and 'anyOf' in items:
            for option in items['anyOf']:
                # If option is a relationship, return it
                if option.get('type') == 'string' and option.get('format') == 'uri':
                    return 'Relationship'

        # If items is directly a uri string type
        if isinstance(items, dict) and items.get('type') == 'string' and items.get('format') == 'uri':
            return 'Relationship'

        # Fallback for arrays
        return 'Property'

    # Format-based DateTime detection (handle case-insensitive)
    if fmt and fmt.lower() in ('date-time', 'datetime', 'date', 'time'):
        return 'DateTime'

    # Default mapping
    mapping = {
        'string': 'Text',
        'integer': 'Number',
        'number': 'Number',
        'boolean': 'Boolean',
        'object': 'StructuredValue'
    }

    return mapping.get(t, 'Property')



def generate_docker_files(path: str = "output_backend"):
    generate_dockerfile(path)
    generate_docker_image(path)
    pass

def generate_dockerfile(path: str):
    with open(path + '/Dockerfile', 'w') as dockerfile:
        dockerfile.write('''FROM python:3.9-slim

# Set working directory
WORKDIR /app

# Create non-root user for security
RUN useradd -m -u 1000 appuser && chown -R appuser:appuser /app

# Copy requirements first for better layer caching
COPY requirements.txt .

# Install dependencies in a single layer
RUN pip install --no-cache-dir -r requirements.txt

# Copy application files
COPY main_api.py pydantic_classes.py sql_alchemy.py ./

# Switch to non-root user
USER appuser

# Expose port
EXPOSE 8000

# Health check
HEALTHCHECK --interval=30s --timeout=3s --start-period=5s --retries=3 \\
    CMD python -c "import requests; requests.get('http://localhost:8000/health', timeout=2)" || exit 1

# Run with uvicorn (production-ready ASGI server)
CMD ["uvicorn", "main_api:app", "--host", "0.0.0.0", "--port", "8000"]
'''
        )
    
    # Generate requirements.txt
    with open(path + '/requirements.txt', 'w') as requirements:
        requirements.write('''fastapi>=0.110.0
uvicorn>=0.28.0
pydantic>=2.6.3
sqlalchemy>=2.0.29
httpx>=0.27.0
requests>=2.31.0
'''
        )
    
    # Generate .dockerignore for smaller build context
    with open(path + '/.dockerignore', 'w') as dockerignore:
        dockerignore.write('''__pycache__
*.pyc
*.pyo
*.pyd
.Python
*.so
*.egg
*.egg-info
dist
build
.git
.gitignore
.env
.venv
venv/
*.log
.pytest_cache
.coverage
htmlcov/
.DS_Store
'''
        )

def generate_docker_image(path: str):
    with open(path + '/create_docker_image.py', 'w') as create_image:
        create_image.write('''import docker
import os
import argparse
import sys

def build_args_parser():
    parser = argparse.ArgumentParser(description='Generating and pushing the image to DockerHub')
    parser.add_argument('-u', '--username',
                        required=True,
                        help='docker username')
    parser.add_argument('-p', '--password',
                        required=True,
                        help='docker password')
    parser.add_argument('-i', '--image',
                        required=True,
                        help='docker image name')
    parser.add_argument('-t', '--tag',
                        required=True,
                        help='tag name')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)
    return parser

def main():
    args = build_args_parser().parse_args()
    # Get dockerfile path
    dockerfile_path = os.path.dirname(os.path.abspath(__file__))
    # Create docker client
    client = docker.from_env()
    # Create docker image
    tag = args.username + '/' + args.image + ':' + args.tag
    image, build_logs = client.images.build(
        path = dockerfile_path,
        rm = True,
        tag = tag
    )

    for log in build_logs:
        print(log)

    client.login(username=args.username, password=args.password)
    resp = client.api.push(
        tag,
        stream=True,
        decode=True,
    )
    
    for line in resp:
        print(line)

if __name__ == '__main__':
    main()
'''
        )


from abc import ABC
from typing import List, Optional, Any, Union
from enum import Enum
from besser.BUML.metamodel.structural import NamedElement, Model


class ControlState(Enum):
    """
    Represents the state of a control qubit.
    """
    CONTROL = 'CONTROL'       # Active on |1> (filled circle)
    ANTI_CONTROL = 'ANTI_CONTROL'  # Active on |0> (empty circle)


# ============================================
# Target Classes (Hardware/Simulator)
# ============================================

class Target(NamedElement, ABC):
    """
    Abstract base class for execution targets.
    """
    def __init__(self, name: str):
        super().__init__(name)


class Hardware(Target):
    """
    Represents a quantum hardware target.
    """
    def __init__(self, name: str):
        super().__init__(name)

    def __repr__(self):
        return f"Hardware(name='{self.name}')"


class Simulator(Target):
    """
    Represents a quantum simulator target.
    """
    def __init__(self, name: str):
        super().__init__(name)

    def __repr__(self):
        return f"Simulator(name='{self.name}')"


# ============================================
# Register and Bit Classes
# ============================================

class QuantumRegister(NamedElement):
    """
    Represents a register of qubits.
    """
    def __init__(self, name: str, size: int):
        super().__init__(name)
        self.size: int = size
        self.qubits: List['Qubit'] = [Qubit(i, i) for i in range(size)]

    def __repr__(self):
        return f"QuantumRegister(name='{self.name}', size={self.size})"


class ClassicalRegister(NamedElement):
    """
    Represents a register of classical bits.
    """
    def __init__(self, name: str, size: int):
        super().__init__(name)
        self.size: int = size
        self.bits: List['ClassicalBit'] = [ClassicalBit(i, i) for i in range(size)]

    def __repr__(self):
        return f"ClassicalRegister(name='{self.name}', size={self.size})"


class Qubit:
    """
    Represents a single qubit.
    """
    def __init__(self, id: int, index: int):
        self.id: int = id
        self.index: int = index

    def __repr__(self):
        return f"Qubit(id={self.id}, index={self.index})"


class ClassicalBit:
    """
    Represents a single classical bit.
    """
    def __init__(self, id: int, index: int):
        self.id: int = id
        self.index: int = index

    def __repr__(self):
        return f"ClassicalBit(id={self.id}, index={self.index})"


# ============================================
# Quantum Circuit
# ============================================

class QuantumCircuit(Model):
    """
    Represents a quantum circuit containing registers and a sequence of operations.
    
    Args:
        name (str): The name of the circuit.
        qubits (int): The number of qubits in the circuit (default register).
    """
    def __init__(self, name: str, qubits: int = 0):
        super().__init__(name)
        self.qregs: List[QuantumRegister] = []
        self.cregs: List[ClassicalRegister] = []
        self.operations: List['QuantumOperation'] = []
        
        # Create a default quantum register if qubits > 0
        if qubits > 0:
            self.add_qreg(QuantumRegister("q", qubits))

    def add_qreg(self, qreg: QuantumRegister):
        """Add a quantum register to the circuit."""
        self.qregs.append(qreg)

    def add_creg(self, creg: ClassicalRegister):
        """Add a classical register to the circuit."""
        self.cregs.append(creg)

    def add_operation(self, operation: 'QuantumOperation'):
        """Add an operation (gate, measurement, etc.) to the circuit."""
        self.operations.append(operation)

    @property
    def num_qubits(self) -> int:
        """Returns the total number of qubits in the circuit."""
        return sum(qreg.size for qreg in self.qregs)

    @property
    def num_clbits(self) -> int:
        """Returns the total number of classical bits in the circuit."""
        return sum(creg.size for creg in self.cregs)

    def __repr__(self):
        return f"QuantumCircuit(name='{self.name}', qregs={self.qregs}, cregs={self.cregs}, ops={len(self.operations)})"


# ============================================
# Quantum Oracle
# ============================================

class QuantumOracle(NamedElement):
    """
    Represents a quantum oracle with a specification.
    """
    def __init__(self, name: str, specification: str):
        super().__init__(name)
        self.specification: str = specification

    def __repr__(self):
        return f"QuantumOracle(name='{self.name}', spec='{self.specification}')"


# ============================================
# Quantum Operation Hierarchy
# ============================================

class QuantumOperation(NamedElement, ABC):
    """
    Abstract base class for all quantum operations.
    """
    def __init__(self, name: str, target_qubits: List[int], control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(name)
        self.target_qubits: List[int] = target_qubits
        self.control_qubits: List[int] = control_qubits if control_qubits else []
        self.control_states: List[ControlState] = control_states if control_states else [ControlState.CONTROL] * len(self.control_qubits)


class Measurement(QuantumOperation):
    """
    Represents a measurement operation.
    """
    def __init__(self, target_qubit: int, output_bit: int = None, basis: str = 'Z'):
        super().__init__(f"Measure_{basis}", [target_qubit])
        self.output_bit: int = output_bit
        self.basis: str = basis  # 'X', 'Y', 'Z'

    def __repr__(self):
        return f"Measurement(qubit={self.target_qubits[0]}, out={self.output_bit}, basis='{self.basis}')"


# ============================================
# Gate Hierarchy
# ============================================

class Gate(QuantumOperation, ABC):
    """
    Abstract base class for unitary quantum gates.
    """
    pass


class GateDefinition(NamedElement):
    """
    Represents the definition of a custom gate.
    """
    def __init__(self, name: str, circuit: 'QuantumCircuit' = None):
        super().__init__(name)
        self.circuit: QuantumCircuit = circuit

    def __repr__(self):
        return f"GateDefinition(name='{self.name}')"


class CustomGate(Gate):
    """
    Represents a user-defined or custom gate.
    
    Args:
        name: The name/label of the custom gate
        target_qubits: List of target qubit indices
        definition: Optional GateDefinition with a full circuit
        gates: Optional list of Gate objects (alternative to definition)
        control_qubits: Optional list of control qubit indices
        control_states: Optional list of control states
    """
    def __init__(self, name: str, target_qubits: List[int], definition: GateDefinition = None, gates: List['Gate'] = None, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(name, target_qubits, control_qubits, control_states)
        self.definition: GateDefinition = definition
        self.gates: List['Gate'] = gates if gates else []

    def __repr__(self):
        return f"CustomGate(name='{self.name}', targets={self.target_qubits}, gates={len(self.gates)})"


class FunctionGate(Gate):
    """
    Represents a function gate defined by a nested circuit.
    """
    def __init__(self, name: str, target_qubits: List[int], definition: GateDefinition = None, gates: List['Gate'] = None, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(name, target_qubits, control_qubits, control_states)
        self.definition: GateDefinition = definition
        self.gates: List['Gate'] = gates if gates else []

    def __repr__(self):
        return f"FunctionGate(name='{self.name}', targets={self.target_qubits}, gates={len(self.gates)})"


class InputGate(Gate):
    """
    Represents input setting gates (InputA, InputB, etc.).
    """
    def __init__(self, input_type: str, target_qubits: List[int], value: str = None):
        super().__init__(input_type, target_qubits)
        self.input_type: str = input_type
        self.value: str = value

    def __repr__(self):
        return f"InputGate(type='{self.input_type}', targets={self.target_qubits}, val={self.value})"


# ============================================
# Abstract Gate Categories
# ============================================

class SingleBitGate(Gate, ABC):
    """
    Abstract base class for single-qubit gates.
    """
    def __init__(self, name: str, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(name, [target_qubit], control_qubits, control_states)


class MultiBitGate(Gate, ABC):
    """
    Abstract base class for multi-qubit gates.
    """
    pass


class ArithmeticGate(Gate, ABC):
    """
    Abstract base class for arithmetic operations.
    """
    def __init__(self, operation_type: str, target_qubits: List[int], input_qubits: List[int] = None, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(operation_type, target_qubits, control_qubits, control_states)
        self.operation_type: str = operation_type
        self.input_qubits: List[int] = input_qubits if input_qubits else []


class ComparisonGate(Gate):
    """
    Represents comparison operations (>, <, ==, etc.).
    """
    def __init__(self, operation: str, target_qubits: List[int], input_qubits: List[int] = None, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(operation, target_qubits, control_qubits, control_states)
        self.operation: str = operation
        self.input_qubits: List[int] = input_qubits if input_qubits else []

    def __repr__(self):
        return f"ComparisonGate(op='{self.operation}', targets={self.target_qubits})"


# ============================================
# Single Bit Gates
# ============================================

class PauliXGate(SingleBitGate):
    """Pauli-X gate (NOT gate)."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("X", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"PauliXGate(target={self.target_qubits[0]})"


class PauliYGate(SingleBitGate):
    """Pauli-Y gate."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("Y", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"PauliYGate(target={self.target_qubits[0]})"


class PauliZGate(SingleBitGate):
    """Pauli-Z gate."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("Z", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"PauliZGate(target={self.target_qubits[0]})"


class HadamardGate(SingleBitGate):
    """Hadamard gate."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("H", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"HadamardGate(target={self.target_qubits[0]})"


class TGate(SingleBitGate):
    """T gate (π/4 phase gate)."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("T", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"TGate(target={self.target_qubits[0]})"


class SGate(SingleBitGate):
    """S gate (π/2 phase gate)."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("S", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"SGate(target={self.target_qubits[0]})"


class PhaseGateZ(SingleBitGate):
    """Z phase gate."""
    def __init__(self, target_qubit: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("Z", target_qubit, control_qubits, control_states)

    def __repr__(self):
        return f"PhaseGateZ(target={self.target_qubits[0]})"


class RXGate(SingleBitGate):
    """Rotation around X-axis gate."""
    def __init__(self, target_qubit: int, theta: float, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("RX", target_qubit, control_qubits, control_states)
        self.theta: float = theta

    def __repr__(self):
        return f"RXGate(target={self.target_qubits[0]}, theta={self.theta})"


class RYGate(SingleBitGate):
    """Rotation around Y-axis gate."""
    def __init__(self, target_qubit: int, theta: float, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("RY", target_qubit, control_qubits, control_states)
        self.theta: float = theta

    def __repr__(self):
        return f"RYGate(target={self.target_qubits[0]}, theta={self.theta})"


class RZGate(SingleBitGate):
    """Rotation around Z-axis gate."""
    def __init__(self, target_qubit: int, theta: float, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("RZ", target_qubit, control_qubits, control_states)
        self.theta: float = theta

    def __repr__(self):
        return f"RZGate(target={self.target_qubits[0]}, theta={self.theta})"


class IdentityGate(SingleBitGate):
    """Identity gate (no operation)."""
    def __init__(self, target_qubit: int):
        super().__init__("I", target_qubit)

    def __repr__(self):
        return f"IdentityGate(target={self.target_qubits[0]})"


# ============================================
# Multi Bit Gates
# ============================================

class CNOT(MultiBitGate):
    """Controlled-NOT (CNOT) gate."""
    def __init__(self, control_qubit: int, target_qubit: int):
        super().__init__("CNOT", [target_qubit], [control_qubit])

    def __repr__(self):
        return f"CNOT(control={self.control_qubits[0]}, target={self.target_qubits[0]})"


class CPhaseGate(MultiBitGate):
    """Controlled phase gate."""
    def __init__(self, control_qubit: int, target_qubit: int, theta: float):
        super().__init__("CP", [target_qubit], [control_qubit])
        self.theta: float = theta

    def __repr__(self):
        return f"CPhaseGate(control={self.control_qubits[0]}, target={self.target_qubits[0]}, theta={self.theta})"


class CYGate(MultiBitGate):
    """Controlled-Y gate."""
    def __init__(self, control_qubit: int, target_qubit: int):
        super().__init__("CY", [target_qubit], [control_qubit])

    def __repr__(self):
        return f"CYGate(control={self.control_qubits[0]}, target={self.target_qubits[0]})"


class CZGate(MultiBitGate):
    """Controlled-Z gate."""
    def __init__(self, control_qubit: int, target_qubit: int):
        super().__init__("CZ", [target_qubit], [control_qubit])

    def __repr__(self):
        return f"CZGate(control={self.control_qubits[0]}, target={self.target_qubits[0]})"


class CXGate(MultiBitGate):
    """Controlled-X gate (same as CNOT)."""
    def __init__(self, control_qubit: int, target_qubit: int):
        super().__init__("CX", [target_qubit], [control_qubit])

    def __repr__(self):
        return f"CXGate(control={self.control_qubits[0]}, target={self.target_qubits[0]})"


class CHGate(MultiBitGate):
    """Controlled-Hadamard gate."""
    def __init__(self, control_qubit: int, target_qubit: int):
        super().__init__("CH", [target_qubit], [control_qubit])

    def __repr__(self):
        return f"CHGate(control={self.control_qubits[0]}, target={self.target_qubits[0]})"


class QFTGate(MultiBitGate):
    """Quantum Fourier Transform gate."""
    def __init__(self, target_qubits: List[int], inverse: bool = False):
        name = "QFT_DAG" if inverse else "QFT"
        super().__init__(name, target_qubits)
        self.inverse: bool = inverse

    def __repr__(self):
        return f"QFTGate(targets={self.target_qubits}, inverse={self.inverse})"


class RXXGate(MultiBitGate):
    """Two-qubit XX rotation gate."""
    def __init__(self, qubit1: int, qubit2: int, theta: float):
        super().__init__("RXX", [qubit1, qubit2])
        self.theta: float = theta

    def __repr__(self):
        return f"RXXGate(qubits={self.target_qubits}, theta={self.theta})"


class RZZGate(MultiBitGate):
    """Two-qubit ZZ rotation gate."""
    def __init__(self, qubit1: int, qubit2: int, theta: float = None):
        super().__init__("RZZ", [qubit1, qubit2])
        self.theta: float = theta

    def __repr__(self):
        return f"RZZGate(qubits={self.target_qubits})"


class CRXGate(MultiBitGate):
    """Controlled RX gate."""
    def __init__(self, control_qubit: int, target_qubit: int, theta: float):
        super().__init__("CRX", [target_qubit], [control_qubit])
        self.theta: float = theta

    def __repr__(self):
        return f"CRXGate(control={self.control_qubits[0]}, target={self.target_qubits[0]}, theta={self.theta})"


class CRYGate(MultiBitGate):
    """Controlled RY gate."""
    def __init__(self, control_qubit: int, target_qubit: int, theta: float):
        super().__init__("CRY", [target_qubit], [control_qubit])
        self.theta: float = theta

    def __repr__(self):
        return f"CRYGate(control={self.control_qubits[0]}, target={self.target_qubits[0]}, theta={self.theta})"


class CRZGate(MultiBitGate):
    """Controlled RZ gate."""
    def __init__(self, control_qubit: int, target_qubit: int, theta: float):
        super().__init__("CRZ", [target_qubit], [control_qubit])
        self.theta: float = theta

    def __repr__(self):
        return f"CRZGate(control={self.control_qubits[0]}, target={self.target_qubits[0]}, theta={self.theta})"


class BellGate(MultiBitGate):
    """Bell state preparation gate."""
    def __init__(self, qubit1: int, qubit2: int):
        super().__init__("Bell", [qubit1, qubit2])

    def __repr__(self):
        return f"BellGate(qubits={self.target_qubits})"


class iSwapGate(MultiBitGate):
    """iSWAP gate."""
    def __init__(self, qubit1: int, qubit2: int):
        super().__init__("iSWAP", [qubit1, qubit2])

    def __repr__(self):
        return f"iSwapGate(qubits={self.target_qubits})"


class SwapGate(MultiBitGate):
    """SWAP gate."""
    def __init__(self, qubit1: int, qubit2: int, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("SWAP", [qubit1, qubit2], control_qubits, control_states)

    def __repr__(self):
        return f"SwapGate(qubits={self.target_qubits})"


class SqrtSwapGate(MultiBitGate):
    """Square root of SWAP gate."""
    def __init__(self, qubit1: int, qubit2: int):
        super().__init__("SqrtSWAP", [qubit1, qubit2])

    def __repr__(self):
        return f"SqrtSwapGate(qubits={self.target_qubits})"


class RYYGate(MultiBitGate):
    """Two-qubit YY rotation gate."""
    def __init__(self, qubit1: int, qubit2: int, theta: float):
        super().__init__("RYY", [qubit1, qubit2])
        self.theta: float = theta

    def __repr__(self):
        return f"RYYGate(qubits={self.target_qubits}, theta={self.theta})"


class RZXGate(MultiBitGate):
    """Two-qubit ZX rotation gate."""
    def __init__(self, qubit1: int, qubit2: int, theta: float):
        super().__init__("RZX", [qubit1, qubit2])
        self.theta: float = theta

    def __repr__(self):
        return f"RZXGate(qubits={self.target_qubits}, theta={self.theta})"


class PhaseGate(SingleBitGate):
    """General Phase gate."""
    def __init__(self, target_qubit: int, angle: float, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__("Phase", target_qubit, control_qubits, control_states)
        self.parameter: float = angle

    def __repr__(self):
        return f"PhaseGate(target={self.target_qubits[0]}, angle={self.parameter})"


class PhaseGradientGate(MultiBitGate):
    """Phase Gradient gate."""
    def __init__(self, target_qubits: List[int], inverse: bool = False):
        name = "PhaseGradient_DAG" if inverse else "PhaseGradient"
        super().__init__(name, target_qubits)
        self.inverse: bool = inverse

    def __repr__(self):
        return f"PhaseGradientGate(targets={self.target_qubits}, inverse={self.inverse})"


class ModularArithmeticGate(ArithmeticGate):
    """Modular Arithmetic gate."""
    def __init__(self, operation: str, target_qubits: List[int], modulo: int, input_qubits: List[int] = None, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(operation, target_qubits, input_qubits, control_qubits, control_states)
        self.modulo: int = modulo

    def __repr__(self):
        return f"ModularArithmeticGate(op='{self.operation_type}', targets={self.target_qubits}, mod={self.modulo})"


class DisplayOperation(QuantumOperation):
    """Display operation (Bloch sphere, Density matrix, etc.)."""
    def __init__(self, display_type: str, target_qubits: List[int]):
        super().__init__(display_type, target_qubits)
        self.display_type: str = display_type

    def __repr__(self):
        return f"DisplayOperation(type='{self.display_type}', targets={self.target_qubits})"


class TimeDependentGate(Gate):
    """Time-dependent gate."""
    def __init__(self, type_name: str, target_qubits: List[int], parameter_expr: str, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(type_name, target_qubits, control_qubits, control_states)
        self.type_name: str = type_name
        self.parameter_expr: str = parameter_expr

    def __repr__(self):
        return f"TimeDependentGate(type='{self.type_name}', targets={self.target_qubits}, expr='{self.parameter_expr}')"


class SpacerGate(Gate):
    """Spacer gate (visual only)."""
    def __init__(self, target_qubits: List[int]):
        super().__init__("Spacer", target_qubits)

    def __repr__(self):
        return f"SpacerGate(targets={self.target_qubits})"


class PostSelection(QuantumOperation):
    """Post-selection operation."""
    def __init__(self, target_qubit: int, value: int, basis: str = 'Z'):
        super().__init__(f"PostSelect_{value}", [target_qubit])
        self.value: int = value
        self.basis: str = basis

    def __repr__(self):
        return f"PostSelection(qubit={self.target_qubits[0]}, val={self.value}, basis='{self.basis}')"


class PrimitiveGate(Gate):
    """Generic primitive gate."""
    def __init__(self, type_name: str, target_qubits: List[int]):
        super().__init__(type_name, target_qubits)
        self.type_name: str = type_name

    def __repr__(self):
        return f"PrimitiveGate(type='{self.type_name}', targets={self.target_qubits})"


class ParametricGate(Gate):
    """Generic parametric gate."""
    def __init__(self, type_name: str, target_qubits: List[int], parameter: float, control_qubits: List[int] = None, control_states: List[ControlState] = None):
        super().__init__(type_name, target_qubits, control_qubits, control_states)
        self.type_name: str = type_name
        self.parameter: float = parameter

    def __repr__(self):
        return f"ParametricGate(type='{self.type_name}', targets={self.target_qubits}, param={self.parameter})"


class OrderGate(Gate):
    """Order gate (for QFT ordering etc)."""
    def __init__(self, order_type: str, target_qubits: List[int]):
        super().__init__(order_type, target_qubits)
        self.order_type: str = order_type

    def __repr__(self):
        return f"OrderGate(type='{self.order_type}', targets={self.target_qubits})"


class ScalarGate(QuantumOperation):
    """Scalar gate (global phase etc)."""
    def __init__(self, scalar_type: str):
        super().__init__(scalar_type, [])
        self.scalar_type: str = scalar_type

    def __repr__(self):
        return f"ScalarGate(type='{self.scalar_type}')"



import pytest
from besser.BUML.metamodel.object import *
from besser.BUML.metamodel.structural import *

########################
#   Structural model   #
########################

attribute1: Property = Property(name="A1", type=IntegerType)
attribute2: Property = Property(name="A2", type=StringType)
class1: Class = Class(name="class1", attributes={attribute1, attribute2})
class2: Class = Class(name="class2", attributes=set())
aend1: Property = Property(name="end1", type=class1, multiplicity=Multiplicity(0, 1))
aend2: Property = Property(name="end2", type=class2, multiplicity=Multiplicity(0, "*"))
association1: BinaryAssociation = BinaryAssociation(name="association1", ends={aend1, aend2})

##########################
#   Object model tests   #
##########################

# Testing model initialization
def test_model_initialization():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    model: ObjectModel = ObjectModel(name="mymodel", objects={obj1,obj2})
    assert len(model.objects) == 2
    model_empty: ObjectModel = ObjectModel(name="mymodel", objects={})
    assert len(model_empty.objects) == 0

# Testing object initialization
def test_object_initialization():
    attr1: AttributeLink = AttributeLink(value=DataValue(classifier=IntegerType, value=100), attribute=attribute1)
    obj1: Object = Object(name="object1", classifier=class1, slots=[attr1])
    assert len(obj1.slots) == 1
    assert obj1.classifier.name == "class1"

# Testing attribute link
def test_attributeLink_initialization():
    attr1: AttributeLink = AttributeLink(value=DataValue(classifier=IntegerType, value=100), attribute=attribute1)
    attr2: AttributeLink = AttributeLink(value=DataValue(classifier=StringType, value="test value"), attribute=attribute2)
    assert attr1.value.value == 100
    assert attr2.value.value == "test value"

# Testing Link inicialization
def test_link_initialization():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    link_end1: LinkEnd = LinkEnd(name="l_end1", association_end=aend1, object=obj1)
    link_end2: LinkEnd = LinkEnd(name="l_end2", association_end=aend2, object=obj2)
    link1: Link = Link(name="link1", association=association1, connections=[link_end1, link_end2])
    assert len(link1.connections) == 2
    assert link1.association.name == "association1"
    assert len(link1.association.ends) == 2

# Testing Link inicialization
def test_linkEnd_initialization():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    link_end1: LinkEnd = LinkEnd(name="l_end1", association_end=aend1, object=obj1)
    link_end2: LinkEnd = LinkEnd(name="l_end2", association_end=aend2, object=obj2)
    assert link_end1.association_end.name == "end1"
    assert link_end2.object.name == "object2"

# Testing link_ends() method
def test_link_ends_method():
    obj1: Object = Object(name="object1", classifier=class1, slots=[])
    obj2: Object = Object(name="object2", classifier=class2, slots=[])
    link_end1: LinkEnd = LinkEnd(name="l_end1", association_end=aend1, object=obj1)
    link_end2: LinkEnd = LinkEnd(name="l_end2", association_end=aend2, object=obj2)
    link1: Link = Link(name="link1", association=association1, connections=[link_end1, link_end2])
    assert len(obj1.link_ends()) == 1
    for link_end in obj1.link_ends():
        assert link_end.object.name == "object2"
    for link_end in obj2.link_ends():
        assert link_end.object.name == "object1"


def test_object_model_validation_passes_for_valid_links():
    worker = Class(name="Worker")
    task = Class(name="Task")
    works_on = Property(name="works_on", type=task, multiplicity=Multiplicity(0, 2))
    assigned_workers = Property(name="assigned_workers", type=worker, multiplicity=Multiplicity(0, 5))
    BinaryAssociation(name="Assignment", ends={works_on, assigned_workers})

    worker_1 = Object(name="worker_1", classifier=worker)
    worker_2 = Object(name="worker_2", classifier=worker)
    task_1 = Object(name="task_1", classifier=task)

    worker_1.works_on = task_1
    worker_2.works_on = task_1

    model = ObjectModel(name="AssignmentModel", objects={worker_1, worker_2, task_1})
    assert model.validate()["success"]


def test_object_model_validation_detects_multiplicity_violation():
    account = Class(name="Account")
    user = Class(name="User")
    owns = Property(name="owns", type=account, multiplicity=Multiplicity(1, 1))
    owner = Property(name="owner", type=user, multiplicity=Multiplicity(1, 1))
    BinaryAssociation(name="Ownership", ends={owns, owner})

    account_a = Object(name="account_a", classifier=account)
    user_alpha = Object(name="user_alpha", classifier=user)
    user_beta = Object(name="user_beta", classifier=user)

    user_alpha.owns = account_a
    user_beta.owns = account_a

    model = ObjectModel(name="OwnershipModel", objects={account_a, user_alpha, user_beta})
    with pytest.raises(ValueError) as excinfo:
        model.validate()
    message = str(excinfo.value)
    assert "multiplicity" in message
    assert "owns" in message or "owner" in message



import os
import subprocess
import sys
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.gui import GUIModel, Module, Button, DataList, DataSourceElement
from besser.BUML.metamodel.structural import DomainModel, PrimitiveDataType, Enumeration
from besser.generators.backend import BackendGenerator
from besser.generators.react import ReactGenerator
from besser.generators import GeneratorInterface
from besser.utilities import sort_by_timestamp

##############################
#   Web Application Generator
##############################
class WebAppGenerator(GeneratorInterface):
    """
    WebAppGenerator is responsible for generating a web application structure based on
    input B-UML and GUI models. It implements the GeneratorInterface and facilitates
    the creation of a web application.

    Args:
        model (DomainModel): The B-UML model representing the application's domain.
        gui_model (GUIModel): The GUI model instance containing necessary configurations.
        output_dir (str, optional): Directory where generated code will be saved. Defaults to None.
    """

    def __init__(self, model: DomainModel, gui_model: GUIModel, output_dir: str = None, agent_model=None):
        super().__init__(model, output_dir)
        self.gui_model = gui_model
        self.agent_model = agent_model
        # Jinja environment configuration
        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        self.env = Environment(loader=FileSystemLoader(templates_path), trim_blocks=True,
                               lstrip_blocks=True, extensions=['jinja2.ext.do'])

    def generate(self):
        """
        Generates web application code based on the provided B-UML and GUI models.
        If agent_model is provided, also generates agent code.

        Returns:
            None, but store the generated code in the specified output directory.
        """
        self._generate_frontend(self.env)
        self._generate_backend(self.env)
        if self.agent_model:
            self._generate_agent(self.env)
        self._generate_docker_files(self.env)

    def _generate_frontend(self, env):
        # Generate frontend code in 'frontend' subfolder
        frontend_dir = os.path.join(self.output_dir, "frontend") if self.output_dir else "frontend"
        frontend_gen = ReactGenerator(self.model, self.gui_model, output_dir=frontend_dir)
        frontend_gen.generate()

    def _generate_backend(self, env):
        # Generate backend code in 'backend' subfolder
        backend_dir = os.path.join(self.output_dir, "backend") if self.output_dir else "backend"
        backend_gen = BackendGenerator(self.model, output_dir=backend_dir)
        backend_gen.generate()
    
    def _generate_agent(self, env):
        """Generate agent code if agent model is provided."""
        from besser.generators.agents.baf_generator import BAFGenerator
        from besser.utilities.buml_code_builder.agent_model_builder import agent_model_to_code
        
        # Generate agent code in 'agent' subfolder
        agent_dir = os.path.join(self.output_dir, "agent") if self.output_dir else "agent"
        os.makedirs(agent_dir, exist_ok=True)
        
        # # Generate agent model file
        # agent_file = os.path.join(agent_dir, "agent_model.py")
        # agent_model_to_code(self.agent_model, agent_file)
        
        # Generate agent files using BAFGenerator
        agent_gen = BAFGenerator(self.agent_model, output_dir=agent_dir)
        agent_gen.generate()

    def _generate_docker_files(self, env):
        """
        Generates Docker-related files for deployment.
        
        Args:
            env: The Jinja2 environment for template rendering.
        """
        # Generate docker-compose.yml
        docker_compose_template = env.get_template('docker-compose.yml.j2')
        docker_compose_path = os.path.join(self.output_dir, 'docker-compose.yml')
        with open(docker_compose_path, 'w') as f:
            f.write(docker_compose_template.render(agent_model=self.agent_model))

        # Generate frontend Dockerfile
        frontend_dockerfile_template = env.get_template('frontend.Dockerfile.j2')
        frontend_dockerfile_path = os.path.join(self.output_dir, 'frontend', 'Dockerfile')
        os.makedirs(os.path.dirname(frontend_dockerfile_path), exist_ok=True)
        with open(frontend_dockerfile_path, 'w') as f:
            f.write(frontend_dockerfile_template.render())

        # Generate backend Dockerfile
        backend_dockerfile_template = env.get_template('backend.Dockerfile.j2')
        backend_dockerfile_path = os.path.join(self.output_dir, 'backend', 'Dockerfile')
        os.makedirs(os.path.dirname(backend_dockerfile_path), exist_ok=True)
        with open(backend_dockerfile_path, 'w') as f:
            f.write(backend_dockerfile_template.render())
        
        # Generate agent Dockerfile if agent model exists
        if self.agent_model:
            agent_dockerfile_template = env.get_template('agent.Dockerfile.j2')
            agent_dockerfile_path = os.path.join(self.output_dir, 'agent', 'Dockerfile')
            os.makedirs(os.path.dirname(agent_dockerfile_path), exist_ok=True)
            with open(agent_dockerfile_path, 'w') as f:
                f.write(agent_dockerfile_template.render(agent_model=self.agent_model))



"""
Domain Model Builder

This module generates Python code for BUML domain models and object models.
"""

import os
from besser.BUML.metamodel.structural.structural import (
    DomainModel,
    AssociationClass,
    Metadata,
    MethodImplementationType,
)
from besser.BUML.metamodel.object.object import ObjectModel
from besser.utilities import sort_by_timestamp as sort
from besser.utilities.buml_code_builder.common import PRIMITIVE_TYPE_MAPPING, safe_class_name


def domain_model_to_code(model: DomainModel, file_path: str, objectmodel: ObjectModel = None):
    """
    Generates Python code for a B-UML model and writes it to a specified file.

    Parameters:
        model (DomainModel): The B-UML model object containing classes, enumerations, 
            associations, and generalizations.
        file_path (str): The path where the generated code will be saved.
        objectmodel (ObjectModel, optional): The B-UML object model to include in the same file.

    Outputs:
        - A Python file containing the base code representation of the B-UML domain model
          and optionally the object model instances.
    """
    output_dir = os.path.dirname(file_path)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not file_path.endswith('.py'):
        file_path += '.py'

    with open(file_path, 'w', encoding='utf-8') as f:
        # Write imports
        f.write("####################\n")
        f.write("# STRUCTURAL MODEL #\n")
        f.write("####################\n\n")
        f.write("from besser.BUML.metamodel.structural import (\n")
        f.write("    Class, Property, Method, Parameter,\n")
        f.write("    BinaryAssociation, Generalization, DomainModel,\n")
        f.write("    Enumeration, EnumerationLiteral, Multiplicity,\n")
        f.write("    StringType, IntegerType, FloatType, BooleanType,\n")
        f.write("    TimeType, DateType, DateTimeType, TimeDeltaType,\n")
        f.write("    AnyType, Constraint, AssociationClass, Metadata, MethodImplementationType\n")
        f.write(")\n")

        # Add object model imports if object model is provided
        if objectmodel:
            f.write("from besser.BUML.metamodel.object import ObjectModel\n")
            f.write("import datetime\n")

        f.write("\n")

        # Write enumerations only if they exist
        enumerations = sort(model.get_enumerations())
        if enumerations:
            f.write("# Enumerations\n")
            for enum in enumerations:
                enum_var_name = safe_class_name(enum.name)
                f.write(f"{enum_var_name}: Enumeration = Enumeration(\n")
                f.write(f"    name=\"{enum.name}\",\n")
                literals_str = ",\n\t\t\t".join([f"EnumerationLiteral(name=\"{lit.name}\")" for lit in sort(enum.literals)])
                f.write(
                    f"    literals={{\n"
                    f"            {literals_str}"
                    f"\n    }}")
                f.write("\n)\n\n")

        # Separate regular classes and association classes
        association_classes = []
        regular_classes = []
        for cls in sort(model.get_classes()):
            if isinstance(cls, AssociationClass):
                association_classes.append(cls)
            else:
                regular_classes.append(cls)

        # Write regular classes
        f.write("# Classes\n")
        for cls in regular_classes:
            cls_var_name = safe_class_name(cls.name)

            # Build class creation parameters
            class_params = [f'name="{cls.name}"']

            if cls.is_abstract:
                class_params.append('is_abstract=True')

            # Add metadata if it exists
            if hasattr(cls, 'metadata') and cls.metadata:
                metadata_params = []
                if cls.metadata.description:
                    # Escape backslashes first, then quotes, then newlines
                    desc = cls.metadata.description.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                    metadata_params.append(f'description="{desc}"')
                if cls.metadata.uri:
                    metadata_params.append(f'uri="{cls.metadata.uri}"')
                if cls.metadata.icon:
                    metadata_params.append(f'icon="{cls.metadata.icon}"')

                if metadata_params:
                    metadata_str = f"Metadata({', '.join(metadata_params)})"
                    class_params.append(f'metadata={metadata_str}')

            f.write(f"{cls_var_name} = Class({', '.join(class_params)})\n")
        f.write("\n")

        # Write class members for regular classes
        for cls in regular_classes:
            cls_var_name = safe_class_name(cls.name)
            f.write(f"# {cls.name} class attributes and methods\n")

            # Write attributes
            for attr in sort(cls.attributes):
                attr_type = PRIMITIVE_TYPE_MAPPING.get(attr.type.name, safe_class_name(attr.type.name))
                visibility_str = f', visibility="{attr.visibility}"' if attr.visibility != "public" else ""
                f.write(f"{cls_var_name}_{attr.name}: Property = Property(name=\"{attr.name}\", "
                       f"type={attr_type}{visibility_str})\n")

            # Write methods
            for method in sort(cls.methods):
                # Extract just the method name (before any parenthesis) for the variable name
                method_var_name = method.name.split('(')[0] if '(' in method.name else method.name
                
                method_type = PRIMITIVE_TYPE_MAPPING.get(method.type.name, safe_class_name(method.type.name)) if method.type else None
                visibility_str = f', visibility="{method.visibility}"' if method.visibility != "public" else ""
                
                # Handle code attribute - escape special characters
                code_str = ""
                if hasattr(method, 'code') and method.code:
                    # Escape backslashes first, then quotes, then newlines
                    escaped_code = method.code.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                    code_str = f', code="{escaped_code}"'

                impl_str = ""
                impl_type = getattr(method, "implementation_type", None)
                if impl_type is not None:
                    impl_name = impl_type.name if hasattr(impl_type, "name") else str(impl_type).upper()
                    impl_str = f", implementation_type=MethodImplementationType.{impl_name}"

                # Build parameters dictionary
                params = {}
                if sort(method.parameters):
                    for param in method.parameters:
                        param_type = PRIMITIVE_TYPE_MAPPING.get(param.type.name, safe_class_name(param.type.name))
                        default_str = f", default_value='{param.default_value}'" if hasattr(param, 'default_value') and param.default_value is not None else ""
                        params[param.name] = f"Parameter(name='{param.name}', type={param_type}{default_str})"

                params_str = "{" + ", ".join(f"{param}" for name, param in params.items()) + "}"

                if method_type:
                    f.write(f"{cls_var_name}_m_{method_var_name}: Method = Method(name=\"{method.name}\""
                           f"{visibility_str}, parameters={params_str}, type={method_type}{code_str}{impl_str})\n")
                else:
                    f.write(f"{cls_var_name}_m_{method_var_name}: Method = Method(name=\"{method.name}\""
                           f"{visibility_str}, parameters={params_str}{code_str}{impl_str})\n")
                if impl_type == MethodImplementationType.STATE_MACHINE:
                    f.write("try:\n")
                    f.write(f"    {cls_var_name}_m_{method_var_name}.state_machine = sm\n")
                    f.write("except NameError:\n")
                    f.write("    pass\n")
                if impl_type == MethodImplementationType.QUANTUM_CIRCUIT:
                    f.write("try:\n")
                    f.write(f"    {cls_var_name}_m_{method_var_name}.quantum_circuit = qc\n")
                    f.write("except NameError:\n")
                    f.write("    pass\n")

            # Write assignments
            if sort(cls.attributes):
                attrs_str = ", ".join([f"{cls_var_name}_{attr.name}" for attr in cls.attributes])
                f.write(f"{cls_var_name}.attributes={{{attrs_str}}}\n")
            if sort(cls.methods):
                # Extract just method names for variable references
                methods_str = ", ".join([f"{cls_var_name}_m_{method.name.split('(')[0] if '(' in method.name else method.name}" for method in cls.methods])
                f.write(f"{cls_var_name}.methods={{{methods_str}}}\n")
            f.write("\n")

        # Get the associations that are not part of association classes
        association_class_associations = [ac.association for ac in association_classes]
        regular_associations = [assoc for assoc in sort(model.associations) if assoc not in association_class_associations]

        # Write regular relationships
        if regular_associations:
            f.write("# Relationships\n")
            for assoc in regular_associations:
                ends_str = []
                for end in sort(assoc.ends):
                    # Determine max value for multiplicity
                    max_value = '"*"' if end.multiplicity.max == "*" else end.multiplicity.max
                    # Use safe class name for the type reference
                    type_var_name = safe_class_name(end.type.name)
                    # Build each property string with all attributes on the same line
                    end_str = (f"Property(name=\"{end.name}\", type={type_var_name}, "
                             f"multiplicity=Multiplicity({end.multiplicity.min}, {max_value})"
                             f"{', is_navigable=' + str(end.is_navigable) if end.is_navigable is not True else ''}"
                             f"{', is_composite=True' if end.is_composite is True else ''})")
                    ends_str.append(end_str)

                # Write the BinaryAssociation with each property on a new line
                f.write(
                    f"{assoc.name}: BinaryAssociation = BinaryAssociation(\n"
                    f"    name=\"{assoc.name}\",\n"
                    "    ends={\n        " + 
                    ",\n        ".join(ends_str) +
                    "\n    }\n"
                    ")\n"
                )
            f.write("\n")

        # Write Association Classes
        if association_classes:
            f.write("# Association Classes\n")
            for ac in association_classes:
                assoc = ac.association
                ac_var_name = safe_class_name(ac.name)
                
                # First write the BinaryAssociation
                ends_str = []
                for end in sort(assoc.ends):
                    # Determine max value for multiplicity
                    max_value = '"*"' if end.multiplicity.max == "*" else end.multiplicity.max
                    # Use safe class name for the type reference
                    type_var_name = safe_class_name(end.type.name)
                    # Build each property string with all attributes on the same line
                    end_str = (f"Property(name=\"{end.name}\", type={type_var_name}, "
                             f"multiplicity=Multiplicity({end.multiplicity.min}, {max_value})"
                             f"{', is_navigable=' + str(end.is_navigable) if end.is_navigable is not True else ''}"
                             f"{', is_composite=True' if end.is_composite is True else ''})")
                    ends_str.append(end_str)

                # Write the BinaryAssociation with each property on a new line
                f.write(
                    f"{assoc.name}: BinaryAssociation = BinaryAssociation(\n"
                    f"    name=\"{assoc.name}\",\n"
                    "    ends={\n        " + 
                    ",\n        ".join(ends_str) +
                    "\n    }\n"
                    ")\n"
                )
                f.write("\n")

                # Now write the attributes for the association class
                for attr in sort(ac.attributes):
                    attr_type = PRIMITIVE_TYPE_MAPPING.get(attr.type.name, safe_class_name(attr.type.name))
                    visibility_str = f', visibility="{attr.visibility}"' if attr.visibility != "public" else ""
                    f.write(f"{ac_var_name}_{attr.name}: Property = Property(name=\"{attr.name}\", "
                           f"type={attr_type}{visibility_str})\n")

                # Write methods for the association class
                for method in sort(ac.methods):
                    # Extract just the method name (before any parenthesis) for the variable name
                    method_var_name = method.name.split('(')[0] if '(' in method.name else method.name
                    
                    method_type = PRIMITIVE_TYPE_MAPPING.get(method.type.name, safe_class_name(method.type.name)) if method.type else None
                    visibility_str = f', visibility="{method.visibility}"' if method.visibility != "public" else ""
                    
                    # Handle code attribute - escape special characters
                    code_str = ""
                    if hasattr(method, 'code') and method.code:
                        # Escape backslashes first, then quotes, then newlines
                        escaped_code = method.code.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                        code_str = f', code="{escaped_code}"'

                    impl_str = ""
                    impl_type = getattr(method, "implementation_type", None)
                    if impl_type is not None:
                        impl_name = impl_type.name if hasattr(impl_type, "name") else str(impl_type).upper()
                        impl_str = f", implementation_type=MethodImplementationType.{impl_name}"

                    # Build parameters dictionary
                    params = {}
                    if sort(method.parameters):
                        for param in method.parameters:
                            param_type = PRIMITIVE_TYPE_MAPPING.get(param.type.name, safe_class_name(param.type.name))
                            default_str = f", default_value='{param.default_value}'" if hasattr(param, 'default_value') and param.default_value is not None else ""
                            params[param.name] = f"Parameter(name='{param.name}', type={param_type}{default_str})"
                    
                    params_str = "{" + ", ".join(f"{param}" for name, param in params.items()) + "}"

                    if method_type:
                        f.write(f"{ac_var_name}_m_{method_var_name}: Method = Method(name=\"{method.name}\""
                               f"{visibility_str}, parameters={params_str}, type={method_type}{code_str}{impl_str})\n")
                    else:
                        f.write(f"{ac_var_name}_m_{method_var_name}: Method = Method(name=\"{method.name}\""
                               f"{visibility_str}, parameters={params_str}{code_str}{impl_str})\n")
                    if impl_type == MethodImplementationType.STATE_MACHINE:
                        f.write("try:\n")
                        f.write(f"    {ac_var_name}_m_{method_var_name}.state_machine = sm\n")
                        f.write("except NameError:\n")
                        f.write("    pass\n")
                    if impl_type == MethodImplementationType.QUANTUM_CIRCUIT:
                        f.write("try:\n")
                        f.write(f"    {ac_var_name}_m_{method_var_name}.quantum_circuit = qc\n")
                        f.write("except NameError:\n")
                        f.write("    pass\n")

                # Create attributes set string if attributes exist
                attributes_str = ""
                if sort(ac.attributes):
                    attrs_str = ", ".join([f"{ac_var_name}_{attr.name}" for attr in ac.attributes])
                    attributes_str = f"attributes={{{attrs_str}}}, "

                # Create methods set string if methods exist
                methods_str = ""
                if sort(ac.methods):
                    # Extract just method names for variable references
                    methods_list = ", ".join([f"{ac_var_name}_m_{method.name.split('(')[0] if '(' in method.name else method.name}" for method in ac.methods])
                    methods_str = f", methods={{{methods_list}}}"

                # Now create the association class
                f.write(f"{ac_var_name} = AssociationClass(\n")
                f.write(f"    name=\"{ac.name}\",\n")
                f.write(f"    {attributes_str}association={assoc.name}{methods_str}\n")
                f.write(")\n\n")
            f.write("\n")

        # Write generalizations
        if model.generalizations:
            f.write("# Generalizations\n")
            for gen in sort(model.generalizations):
                specific_var_name = safe_class_name(gen.specific.name)
                general_var_name = safe_class_name(gen.general.name)
                f.write(f"gen_{gen.specific.name}_{gen.general.name} = Generalization"
                        f"(general={general_var_name}, specific={specific_var_name})\n")
            f.write("\n")

        # Write OCL constraints if they exist
        if hasattr(model, 'constraints') and model.constraints:
            f.write("\n# OCL Constraints\n")
            for constraint in sort(model.constraints):
                constraint_name = constraint.name.replace("-", "_")
                context_var_name = safe_class_name(constraint.context.name)
                f.write(f"{constraint_name}: Constraint = Constraint(\n")
                f.write(f"    name=\"{constraint.name}\",\n")
                f.write(f"    context={context_var_name},\n")
                f.write(f"    expression=\"{constraint.expression}\",\n")
                f.write(f"    language=\"{constraint.language}\"\n")
                f.write(")\n")
            f.write("\n")

        # Write domain model
        f.write("# Domain Model\n")
        
        # Write domain model metadata if it exists
        domain_metadata_var = None
        if hasattr(model, 'metadata') and model.metadata:
            domain_metadata_var = "domain_metadata"
            f.write(f"{domain_metadata_var} = Metadata(\n")
            if model.metadata.description:
                # Escape quotes and newlines in description
                desc = model.metadata.description.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                f.write(f'    description="{desc}",\n')
            if model.metadata.uri:
                f.write(f'    uri="{model.metadata.uri}",\n')
            if model.metadata.icon:
                f.write(f'    icon="{model.metadata.icon}"\n')
            f.write(")\n\n")
        
        f.write("domain_model = DomainModel(\n")
        f.write(f"    name=\"{model.name}\",\n")

        # Include all classes (regular and association) and enumerations in types
        class_names = ', '.join(safe_class_name(cls.name) for cls in sort(model.get_classes()))
        enum_names = ', '.join(safe_class_name(enum.name) for enum in sort(model.get_enumerations()))
        types_str = (f"{class_names}, {enum_names}" if class_names and enum_names else
                    class_names or enum_names)
        f.write(f"    types={{{types_str}}},\n")

        # Include both regular associations and those used in association classes
        all_assoc_names = ', '.join([assoc.name for assoc in regular_associations] + 
                                    [ac.association.name for ac in association_classes])
        if all_assoc_names:
            f.write(f"    associations={{{all_assoc_names}}},\n")
        else:
            f.write("    associations={},\n")

        if hasattr(model, 'constraints') and model.constraints:
            constraints_str = ', '.join(c.name.replace("-", "_") for c in sort(model.constraints))
            f.write(f"    constraints={{{constraints_str}}},\n")
        if model.generalizations:
            f.write(f"    generalizations={{{', '.join(f'gen_{gen.specific.name}_{gen.general.name}' for gen in sort(model.generalizations))}}},\n")
        else:
            f.write("    generalizations={},\n")
        
        # Add metadata if it exists
        if domain_metadata_var:
            f.write(f"    metadata={domain_metadata_var}\n")
        else:
            f.write("    metadata=None\n")
        f.write(")\n")

        # Generate object model code if provided
        if objectmodel:
            f.write("\n################\n")
            f.write("# OBJECT MODEL #\n")
            f.write("################\n")

            # Write object instances using fluent API
            for obj in sorted(objectmodel.objects, key=lambda x: x.name_):
                obj_var_name = f"{obj.name_.lower()}_obj"
                classifier_var_name = safe_class_name(obj.classifier.name)

                # Start the fluent API call using the proper syntax: Class("name")
                f.write(f"{obj_var_name} = {classifier_var_name}(\"{obj.name_}\")")

                # Add attributes if the object has slots
                if obj.slots:
                    attributes_dict = {}
                    for slot in obj.slots:
                        attr_name = slot.attribute.name

                        # Format the value based on type
                        if isinstance(slot.value.value, str):
                            value_str = f'"{slot.value.value}"'
                        elif hasattr(slot.value.value, 'isoformat'):  # datetime objects
                            value_str = f'datetime.datetime.fromisoformat("{slot.value.value.isoformat()}")'
                        elif hasattr(slot.value.value, 'owner') and hasattr(slot.value.value.owner, 'name'):
                            # This is an enumeration literal - generate the proper reference
                            enum_name = slot.value.value.owner.name
                            literal_name = slot.value.value.name
                            value_str = f"{enum_name}.{literal_name}"
                        else:
                            value_str = str(slot.value.value)

                        attributes_dict[attr_name] = value_str

                    # Add attributes to the fluent API call
                    if attributes_dict:
                        attr_pairs = [f"{k}={v}" for k, v in attributes_dict.items()]
                        f.write(f".attributes({', '.join(attr_pairs)})")

                # Complete the fluent API call
                f.write(".build()\n")

            f.write("\n")

            # Add links after objects are created (avoiding forward reference issues)
            if hasattr(objectmodel, 'links') and objectmodel.links:

                # Group links by (source_obj_var, end_name)
                grouped_links = {}
                for link in objectmodel.links:
                    if len(link.connections) == 2:
                        end1, end2 = link.connections
                        obj1_var = f"{end1.object.name_.lower()}_obj"
                        obj2_var = f"{end2.object.name_.lower()}_obj"
                        end2_name = end2.association_end.name

                        key = (obj1_var, end2_name)
                        grouped_links.setdefault(key, set()).add(obj2_var)

                # Write assignments for each group
                for (obj_var, end_name), targets in grouped_links.items():
                    if len(targets) == 1:
                        [single_target] = targets
                        f.write(f"{obj_var}.{end_name} = {single_target}\n")
                    else:
                        target_str = ", ".join(sorted(targets))  # sorted for consistency
                        f.write(f"{obj_var}.{end_name} = {{{target_str}}}\n")

                f.write("\n")

            # Create the object model instance
            f.write("# Object Model instance\n")
            objects_str = ", ".join([f"{obj.name_.lower()}_obj" for obj in sorted(objectmodel.objects, key=lambda x: x.name_)])
            f.write(f"object_model: ObjectModel = ObjectModel(\n")
            f.write(f"    name=\"{objectmodel.name}\",\n")
            f.write(f"    objects={{{objects_str}}}")
            
            # Add metadata if it exists
            if hasattr(objectmodel, 'metadata') and objectmodel.metadata:
                if objectmodel.metadata.description:
                    # Escape quotes and newlines in description
                    desc = objectmodel.metadata.description.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                    f.write(",\n")
                    f.write(f'    metadata=Metadata(description="{desc}")\n')
                else:
                    f.write("\n")
            else:
                f.write("\n")
            f.write(")\n")

    print(f"BUML model saved to {file_path}")
